module lkuForm.

%% The abstract datatype for formulas in LKU.

% Constants: list all eight (LKU paper lists only four)
type  t+1, t+2, t-1, t-2, f+1, f+2, f-1, f-2 form.

% Connectives
type &+2&, !+2!	 	form -> form -> form.
type &+1&, !+1! 	form -> form -> form.
type &-1&, !-1! 	form -> form -> form.
type &-2&, !-2!		form -> form -> form.

% Quantifiers
type all1, all2, some1, some2	(A -> form ) -> form.	

% Priority for connectives
infixr  &+1&, &+2&, &-1&, &-2&        5.
infixr	!+1!, !+2!, !-1!, !-2!        4.

% Attributing atoms a polarity
type p1, p2, n2, n1             atm -> form.

% Negate a formula by computing the de Morgan dual.
negate (A &+1& B) (A' !-1! B') & negate (A &-1& B) (A' !+1! B') &
negate (A !-1! B) (A' &+1& B') & negate (A !+1! B) (A' &-1& B') &
negate (A !-2! B) (A' &+2& B') & negate (A !+2! B) (A' &-2& B') &
negate (A &+2& B) (A' !-2! B') & negate (A &-2& B) (A' !+2! B') :- 
   negate A A', negate B B'.

negate (all1 A) (some1 A') & negate (some1 A) (all1 A') &
negate (all2 A) (some2 A') & negate (some2 A) (all2 A') :- 
   pi x\ negate (A x) (A' x).

negate t+1 f-1 & negate t+2 f-2 & negate t-1 f+1 & negate t-2 f+2.
negate f-1 t+1 & negate f-2 t+2 & negate f+1 t-1 & negate f+2 t-2 .

negate (p1 A) (n1 A) & negate (n1 A) (p1 A).
negate (p2 A) (n2 A) & negate (n2 A) (p2 A).

pos1_lit (p1 _) & pos2_lit (p2 _).
neg1_lit (n1 _) & neg2_lit (n2 _).
pos_lit P :- pos1_lit P ; pos2_lit P.
neg_lit P :- neg1_lit P ; neg2_lit P.

pos1_form (p1 _) & pos1_form (_ &+1& _) & pos1_form (_ !+1! _).
pos1_form (some1 _) & pos1_form t+1 & pos1_form f+1.
pos2_form (p2 _) & pos2_form (_ &+2& _) & pos2_form (_ !+2! _).
pos2_form (some2 _) & pos2_form t+2 & pos2_form f+2.
neg1_form (n1 _) & neg1_form (_ !-1! _) & neg1_form (_ &-1& _).
neg1_form (all1 _) & neg1_form t-1 & neg1_form f-1.
neg2_form (n2 _) & neg2_form (_ &-2& _) & neg2_form (_ !-2! _).
neg2_form (all2 _) & neg2_form t-2 & neg2_form f-2.

isPos F :- pos1_form F; pos2_form F.
isNeg F :- neg1_form F; neg2_form F.
litOrPos C :- neg_lit C; isPos C.

composed_form F :- (conj+2 _ _ F). 
composed_form F :- (conj+1 _ _ F). 
composed_form F :- (conj-1 _ _ F). 
composed_form F :- (conj-2 _ _ F). 
composed_form F :- (disj+2 _ _ F). 
composed_form F :- (disj+1 _ _ F). 
composed_form F :- (disj-1 _ _ F). 
composed_form F :- (disj-2 _ _ F). 


% constructor / destructor for this abstract data type

conj+1 B C (B &+1& C) & conj+2 B C (B &+2& C).
conj-1 B C (B &-1& C) & conj-2 B C (B &-2& C).
conj-  B C Conj :- conj-1 B C Conj; conj-2 B C Conj.
conj+  B C Conj :- conj+1 B C Conj; conj+2 B C Conj.
disj+1 B C (B !+1! C) & disj+2 B C (B !+2! C).
disj-1 B C (B !-1! C) & disj-2 B C (B !-2! C).
disj-  B C Conj :- disj-1 B C Conj; disj-2 B C Conj.
disj+  B C Conj :- disj+1 B C Conj; disj+2 B C Conj.
true+1 t+1 & true+2 t+2 & true-1 t-1 & true-2 t-2.
true+ B :- true+1 B ; true+2 B.
true- B :- true-1 B ; true-2 B.
false+1 f+1 & false+2 f+2 & false-1 f-1 & false-2 f-2.
false+ B :- false+1 B ; false+2 B.
false- B :- false-1 B ; false-2 B.
all-1  B (all1 B) & all-2  B (all2 B).
all-   B Q :- all-1 B Q ; all-2 B Q.
some+1 B (some1 B) & some+2 B (some2 B).
some+  B Q :- some+1 B Q ; some+2 B Q.
lit-1 A (n1 A) & lit-2 A (n2 A) & lit+1 A (p1 A) & lit+2 A (p2 A).

ensure+2 B B :- pos2_form B.
ensure+2 B C :- (isNeg B ; pos1_form B), conj+2 B t+2 C.
ensure+1 B B :- pos1_form B.
ensure+1 B C :- (isNeg B ; pos2_form B), conj+1 B t+1 C.
ensure-2 B B :- neg2_form B.
ensure-2 B C :- (isPos B ; neg1_form B), conj-2 B t-2 C.
ensure-1 B B :- neg1_form B.
ensure-1 B C :- (isPos B ; neg2_form B), conj-1 B t-1 C.
