module stlc-fpc.
accumulate lists. 


/********************************/
/*        Experts               */
/********************************/

/* If we have a delayed formula we always decide on it, unless there is
  a redex in the certificate */
decideExpert (d_idx F)
	(st_term_cert T D I A)
	(st_term_cert T D I A)
	(boundry MSet) :-
	mset_singleton (d_idx F) MSet,
	not_redex T.
decideExpert (d_idx F)
	(st_args_cert Args D I A)
	(st_args_cert Args D I A)
	(boundry MSet) :-
	mset_singleton (d_idx F) MSet.
decideExpert (d_idx F)
	(st_extra_cert I A)
	(st_extra_cert I A)
	(boundry MSet) :-
	mset_singleton (d_idx F) MSet.


/* If we only have one variable with a non-functional type then we decide 
  on it and then we are done */
decideExpert (c_idx PosLit)
	(st_term_cert (var J) D I []) 
	(st_empty_cert)
	(boundry MSet) :- 
	PosLit is D - J - 1,
	czone (c_idx PosLit),
	ref (c_idx PosLit) F,
	pos1_lit F.


/* If that variable has a functional type (i.e. the index refers to a tensor
  formula) this means that the term is not in eta long normal form.
  We thus continue with the list of extra arguments already stored
  and the current index if we are going to find more eta redexes (and thus
  more extra arguments to store).  */
decideExpert (c_idx PosLit)
	(st_term_cert (var J) D I A) 
	(st_extra_cert I A)
	(boundry MSet) :- 
	PosLit is D - J - 1,
	czone (c_idx PosLit),
	ref (c_idx PosLit) F,
	conj+1 _ _ F.

/* If we only have one head variable applied to a list of arguments then
  we decide on the variable and switch the certificate to indicate that
  we are checking a list of arguments (corresponding to a backbone of tensors) */
decideExpert (c_idx PosLit) 
	(st_term_cert ((var J) app U) D I A) 
	(st_args_cert U D I A) 
	(boundry MSet) :- 
	PosLit is D - J - 1.


/* If we are checking a subformula involving eta-redexes, then we decide on the 
  first formula stored in the list of extra arguments. We continue with the 
  tail of that list. */
decideExpert (x_idx PosLit)
	(st_extra_cert I (PosLit::L))
	(st_extra_cert I L)
	(boundry MSet).



/* When checking a backbone of tensors, we:
  - check the first argument of the list with an empty linear zone 
  - continue the checking of the list, passing the linear zone 
    which consists of only one element
 We have three cases:
 - there is at least two arguments in arguments of the lambda term
    we thus put the first one on the left (with an empty list of extra args)
    and the rest on the right
 - there is only one argument remaining  in the lambda term
   we put this argument to the left and all the extra arguments to the right
 - there are only extra arguments
    we thus put the first one on the left and the rest on the right 
 */
andExpert Empty DeltaR 
	(st_args_cert (H::X::T) D I A) 
	(st_term_cert H D I [])
	(st_args_cert (X::T) D I A)
	(sync Delta F) :-
	conj+1 _ _ F,
	mset_empty Empty,
	mset_eq DeltaR Delta.

andExpert Empty DeltaR 
	(st_args_cert [H] D I A) 
	(st_term_cert H D I [])
	(st_extra_cert I A)
	(sync Delta F) :-
	conj+1 _ _ F,
	mset_empty Empty,
	mset_eq DeltaR Delta.

andExpert Empty Delta
	(st_extra_cert I (H::A))
	(st_extra_cert I [H])
	(st_extra_cert I A)
	(sync Delta F) :-
	conj+1 _ _ F,
        mset_empty Empty.


/* The trivial case to handle delays */
andExpert Empty Delta
	Cert
	Cert 
	Cert
	(sync Delta F) :-
	conj+2 _ _ F,
	mset_empty Empty.


cutExpert C Delta Empty 
	(st_term_cert (redex_t (db_abs T) U Type) D I A)
	(st_term_cert (db_abs T) D I A)
	(st_term_cert U D I A)
	(boundry Delta) :-
	st_embed_form_l_lku Type C.


existsExpert _ X X _.
releaseExpert X X _.

/* We would like to discard cases where we could success whereas the 
  information provided is not correct. For instance we could accept:
ex_counter 5
	(db_abs ( db_abs ( redex_t
				(db_abs ( redex_t (db_abs (var 1)) (var 0)
						((st_atm 0) st_arr (st_atm 0))
				))
				(var 1)
				(st_atm 0))))
	((st_atm 0) st_arr ((st_atm 0) st_arr (st_atm 0))) .
whereas ((st_atm 0) st_arr (st_atm 0)) is not the correct subtype */
initialExpert _ (st_empty_cert) _.
initialExpert _ (st_extra_cert _ []) _.
trueExpert _ _.
		

/********************************/
/*     Computation devices      */
/********************************/


/* Pars
 We discard the case where we have a redex.
 We could simply have:
    orClerk F C C :- disj-1 _ _ F.
 but doing so speeds up things by avoiding backtracking */
orClerk F (st_term_cert (db_abs T) D I A) (st_term_cert (db_abs T) D I A) :-
	disj-1 _ _ F.	
orClerk F (st_term_cert (var J) D I A) (st_term_cert (var J) D I A) :-	
	disj-1 _ _ F.	
orClerk F (st_term_cert (U app V) D I A) (st_term_cert (U app V) D I A) :-	
	disj-1 _ _ F.	
orClerk F (st_extra_cert I A) (st_extra_cert I A) :-
	disj-1 _ _ F.	


/* The counterpart of the postive delay */
orClerk F C C :- disj-2 _ _ F.


/* When storing in classical zone occurs, there are four possibilities:
  - there is an abstraction at the top of the lambda term
     this is the usual case where we increment the depth
  - there is a variable or an application
     this means there is an eta redex, we thus add the current index
     to the list of extra arguments
  - we are already checking a subformula involving an eta-redex
     (the lambda term is already entirely read). We thus also
     add the index to the list of extra arguments
  - there is a redex at the top. This means that there is an interaction 
    between beta and eta. We choose to fail and thus the cut will apply. 
    Doing so we we store types as they appear in the lambda term 
    (otherwise we could permute them).
*/ 


store2Clerk _ (c_idx D)
	(st_term_cert (db_abs T) D I A) 
	(st_term_cert T D' I A) :- 
	D' is D + 1.

store2Clerk _ (x_idx I)
	(st_term_cert (var J) D I A) 
	(st_term_cert (var J) D I' AI) :- 
	I' is I + 1,
	append A [I] AI.
store2Clerk _ (x_idx I)
	(st_term_cert (U app V) D I A) 
	(st_term_cert (U app V) D I' AI) :- 
	I' is I + 1,
	append A [I] AI.

store2Clerk _ (x_idx I) 
	(st_extra_cert I A) 
	(st_extra_cert I' AI) :- 
	I' is I + 1,
	append A [I] AI. 


/* Store a literal */
store1Clerk F l_idx C C :- neg1_lit F.

/* Store a delayed formula */
store1Clerk F (d_idx F) C C :- conj+2 _ _ F.

allClerk _ X X.

/* Useful since false-2 can be introduced as we have true+2 delays */
falseClerk _ X X. 

/* Useless since neither all and false connectives nor async conj exist.
all_state _ X X.		
false_state _ X X.
conj_state _ X X X. */


not_redex (var _).
not_redex (db_abs _).
not_redex (_ app _).
